/* Name: main.c
 * Project: HID-Test
 * Author: Christian Starkjohann
 * Creation Date: 2006-02-02
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 299 2007-03-29 17:07:19Z cs $
 */

#ifndef F_CPU
	#define  F_CPU   12000000L    /* 12 MHz */
#endif


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"






/* --------------------------------- RC5 ---------------------------------- */

#define uchar unsigned char
#define uint unsigned int

#define	xRC5_IN		PIND
#define	xRC5		PD3			// IR input low active

#define RC5TIME 	1.778e-3	// 1.778msec
// f= F_CPU/Vorteiler = F_CPU/64--> T= 64/F_CPU = Zeit fr 1 Zyklus
// Timer vorgeladen mit -6=250 --> 6 Zyklen bis Interrupt
// --> T= 6*64/F_CPU = 384/F_CPU
// Anzahl der Interrupt bis RC5TIME vorbei 
// --> PULSE= RC5TIME/(384/F_CPU) = (RC5TIME/384)*F_CPU = F_CPU/384*RC5TIME

#define PULSE_MIN	(uchar)(F_CPU / 384 * RC5TIME * 0.4 + 0.5)
#define PULSE_1_2	(uchar)(F_CPU / 384 * RC5TIME * 0.8 + 0.5)
#define PULSE_MAX	(uchar)(F_CPU / 384 * RC5TIME * 1.2 + 0.5)


volatile uchar	rc5_bit;				// bit value
volatile uchar	rc5_time;				// count bit time
volatile uint	rc5_tmp;				// shift bits in
volatile uint	rc5_data;				// store result
volatile uint 	toggle=2;				// toggle bit




static uchar RC5( void )
{
  uint i;
  uchar out=0;

  i = rc5_data;					// 2 bytes von interrupt lesen
  rc5_data = 0;
  if( i )						// | 1 |-C6| T | A4| A3| A2| A1| A0| C5| C4| C3| C2| C1| C0|
  {
	//11 nach rechts, mit 1 UND:
	//  | 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 1| -C6| T|
	// &| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0  | 1|
	// =| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0  | T|
    //itoa( i >> 6 & 0x1F, s, 10);					// Device address
	// 6 nach rechts
	//  | 0 | 0 | 0 | 0 | 0 | 0 | 1 |-C6| T | A4| A3| A2| A1| A0|
	// &| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 1 | 1 | 1 | 1 |
	// =| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | A4| A3| A2| A1| A0|
    
	//itoa((i & 0x3F) | (~i >> 6 & 0x40), s, 10);	// Key Code
	//  | 1 |-C6| T | A4| A3| A2| A1| A0| C5| C4| C3| C2| C1| C0|
	// &| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 1 | 1 | 1 | 1 | 1 |
	// =| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | C5| C4| C3| C2| C1| C0|
	//
	//  | 0 | C6|-T |-A4|-A3|-A2|-A1|-A0|-C5|-C4|-C3|-C2|-C1|-C0|
	// 6 Stellen nach rechts verschieben
	//  | 0 | 0 | 0 | 0 | 0 | 0 | 0 | C6|-T |-A4|-A3|-A2|-A1| A0|
	// &| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0 |
	// =| 0 | 0 | 0 | 0 | 0 | 0 | 0 | C6| 0 | 0 | 0 | 0 | 0 | 0 |
	//
	//  | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | C5| C4| C3| C2| C1| C0|
	// v| 0 | 0 | 0 | 0 | 0 | 0 | 0 | C6| 0 | 0 | 0 | 0 | 0 | 0 |
	// =| 0 | 0 | 0 | 0 | 0 | 0 | 0 | C6| C5| C4| C3| C2| C1| C0|

	//if ((( i >> 11 & 1)==toggle) && ((i >> 6 & 0x1F)==4)) {
	
	if ((i >> 6 & 0x1F)==5) {
	  switch( ((i & 0x3F) | (~i >> 6 & 0x40)) )	{
	  	case 80: {	//hoch
			out=1;
			break;
		}
	  	case 81: {	//runter
			out=2;
			break;
		}
	  	case 85: {	//links
			out=3;
			break;
		}
	  	case 86: {	//rechts
			out=4;
			break;
		}
	  	case 87: {	//OK
			out=5;
			break;
		}
	  	case 32: {	//CH+
			out=6;
			break;
		}
	  	case 33: {	//CH-
			out=7;
			break;
		}
	  	case 37: {	//rewind
			out=10;
			break;
		}
	  	case 53: {	//play
			out=11;
			break;
		}
	  	case 38: {	//forward
			out=12;
			break;
		}
	  	case 54: {	//stop
			out=13;
			break;
		}
	  	case 41: {	//pause
			out=14;
			break;
		}
	  	case 64: {	//videotext
			out=15;
			break;
		}
	  	case 12: {	//ausschalten
			out=16;
			break;
		}
		default: { out=0;}
	  } // switch
	} // if

	if ((i >> 6 & 0x1F)==0) {
	  switch( ((i & 0x3F) | (~i >> 6 & 0x40)) )	{
	  	case 16: {	//Vol+
			out=8;
			break;
		}
	  	case 17: {	//Vol-
			out=9;
			break;
		}
	  } // switch
	} // if

	toggle=( i >> 11 & 1);	
/*
	DDRD|=(1<<PD5);
	PORTD^=(1<<PD5);*/	
  }
  return out;	
}

/* ----------------------- hardware I/O abstraction ------------------------ */

/* pin assignments:
PB0	Key 1
PB1	Key 2
PB2	Key 3
PB3	Key 4
PB4	Key 5
PB5 Key 6

PC0	Key 7
PC1	Key 8
PC2	Key 9
PC3	Key 10
PC4	Key 11
PC5	Key 12

PD0	USB-
PD1	debug tx
PD2	USB+ (int0)
PD3	Key 13
PD4	Key 14
PD5	Key 15
PD6	Key 16
PD7	Key 17
*/

static void hardwareInit(void)
{
uchar	i, j;
/*
    PORTB = 0xff;   // activate all pull-ups 
    DDRB = 0;       // all pins input 
    PORTC = 0xff;   // activate all pull-ups 
    DDRC = 0;       // all pins input 
    PORTD = 0xfa;   // 1111 1010 bin: activate pull-ups except on USB lines 
    DDRD = 0x07;    // 0000 0111 bin: all pins input except USB (-> USB reset) 
*/
	j = 0;
	while(--j){     /* USB Reset by device only required on Watchdog Reset */
		i = 0;
		while(--i); /* delay >10ms for USB reset */
	}
//    DDRD = 0x02;    /* 0000 0010 bin: remove USB reset condition */
    /* configure timer 0 for a rate of 12M/(1024 * 256) = 45.78 Hz (~22ms) */
    TCCR0 = 5;      /* timer 0 prescaler: 1024 */
}

/* ------------------------------------------------------------------------- */

#define NUM_KEYS    16

/* The following function returns an index for the first key pressed. It
 * returns 0 if no key is pressed.
 */
static uchar    keyPressed(void)
{
uchar   i, mask, x;

    x = PINB;
    mask = 1;
    for(i=0;i<6;i++){
        if((x & mask) == 0)
            return i + 1;
        mask <<= 1;
    }
    x = PINC;
    mask = 1;
    for(i=0;i<6;i++){
        if((x & mask) == 0)
            return i + 7;
        mask <<= 1;
    }
    x = PIND;
    mask = 1 << 3;
    for(i=0;i<5;i++){
        if((x & mask) == 0)
            return i + 13;
        mask <<= 1;
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

PROGMEM char usbHidReportDescriptor[35] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

#define KEY_A       4
#define KEY_B       5
#define KEY_C       6
#define KEY_D       7
#define KEY_E       8
#define KEY_F       9
#define KEY_G       10
#define KEY_H       11
#define KEY_I       12
#define KEY_J       13
#define KEY_K       14
#define KEY_L       15
#define KEY_M       16
#define KEY_N       17
#define KEY_O       18
#define KEY_P       19
#define KEY_Q       20
#define KEY_R       21
#define KEY_S       22
#define KEY_T       23
#define KEY_U       24
#define KEY_V       25
#define KEY_W       26
#define KEY_X       27
#define KEY_Y       28
#define KEY_Z       29
#define KEY_1       30
#define KEY_2       31
#define KEY_3       32
#define KEY_4       33
#define KEY_5       34
#define KEY_6       35
#define KEY_7       36
#define KEY_8       37
#define KEY_9       38
#define KEY_0       39

#define KEY_F1      58
#define KEY_F2      59
#define KEY_F3      60
#define KEY_F4      61
#define KEY_F5      62
#define KEY_F6      63
#define KEY_F7      64
#define KEY_F8      65
#define KEY_F9      66
#define KEY_F10     67
#define KEY_F11     68
#define KEY_F12     69

#define KEY_ENTER   40
#define KEY_SPACE   44
#define KEY_RARROW  79
#define KEY_LARROW  80
#define KEY_DOWN	81
#define KEY_UP		82


static const uchar  keyReport[NUM_KEYS + 1][2] PROGMEM = {
/* none */  {0, 0},                     /* no key pressed */
/*  1 */    {MOD_CONTROL_RIGHT, KEY_RARROW},//hoch
/*  2 */    {MOD_CONTROL_RIGHT, KEY_LARROW},//runter
/*  3 */    {MOD_SHIFT_RIGHT, KEY_LARROW},	//links
/*  4 */    {MOD_SHIFT_RIGHT, KEY_RARROW},	//rechts
/*  5 */    {0, KEY_ENTER},					//OK
/*  6 */    {0, KEY_N},						//CH+
/*  7 */    {0, KEY_P},						//CH-
/*  8 */    {MOD_CONTROL_LEFT, KEY_UP},		//Vol+
/*  9 */    {MOD_CONTROL_LEFT, KEY_DOWN},	//Vol-
/* 10 */    {0, KEY_P},						//rewind
/* 11 */    {0, KEY_SPACE},					//play
/* 12 */    {0, KEY_N},						//forward
/* 13 */    {0, KEY_S},						//stop
/* 14 */    {0, KEY_SPACE},					//pause
/* 15 */    {MOD_ALT_RIGHT, KEY_D},			//videotext, DualView
/* 16 */    {MOD_ALT_LEFT, KEY_F4},			//ausschalten

};

static void buildReport(uchar key)
{
/* This (not so elegant) cast saves us 10 bytes of program memory */
    *(int *)reportBuffer = pgm_read_word(keyReport[key]);
}

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport(keyPressed());
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

/* ------------------------------------------------------------------------- */

int	main(void)
{
uchar   key, lastKey = 0, keyDidChange = 0;
uchar   idleCounter = 0;

	wdt_enable(WDTO_2S);
    hardwareInit();
	odDebugInit();
	usbInit();


	// Timer fr RC5
  	TCCR2 = 1<<CS22;			//Teiler 64
	TCNT2 = -6;
  	TIMSK |= 1<<TOIE2;			//timer interrupt aktivieren

	sei();
    DBG1(0x00, 0, 0);
	for(;;){	/* main event loop */
		wdt_reset();
		usbPoll();
        //key = keyPressed();
		key = RC5();
        if(lastKey != key){
            lastKey = key;
            keyDidChange = 1;
        }
        if(TIFR & (1<<TOV0)){   /* 22 ms timer */
            TIFR = 1<<TOV0;
            if(idleRate != 0){
                if(idleCounter > 4){
                    idleCounter -= 5;   /* 22 ms in units of 4 ms */
                }else{
                    idleCounter = idleRate;
                    keyDidChange = 1;
                }
            }
        }
        if(keyDidChange && usbInterruptIsReady()){
            keyDidChange = 0;
            /* use last key and not current key status in order to avoid lost
               changes in key status. */
            buildReport(lastKey);
            usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        }
	}
	return 0;
}

/* ------------------------------------------------------------------------- */



ISR (TIMER2_OVF_vect)
{
  uint tmp = rc5_tmp;					// lakale variable hat schnelleren "zugang"
  TCNT2 = -6;							// 2 * 256 = 512 Zyklen, (-2)-->254
  // f=MHz/Vorteiler=12000000/64=187500
  // 1/187500=5,33s= 1 Zyklus
  // da mit 250 vorgeladen--> 6 Zyklen --> alle 32s ein Interrupt

  if( ++rc5_time > PULSE_MAX )				// zhlt pulse time
  {				
    if( !(tmp & 0x4000) && tmp & 0x2000 )	// nur wenn 14 bits empfangen
      rc5_data = tmp;
    tmp = 0;
  }

  if( (rc5_bit ^ xRC5_IN) & 1<<xRC5 )		// change detect, ^ -->EXOR
  {		
    rc5_bit = ~rc5_bit;						// 0x00 -> 0xFF -> 0x00

    if( rc5_time < PULSE_MIN )				// zu kurz
      tmp = 0;

    if( !tmp || rc5_time > PULSE_1_2 )		// start oder lange pulse time
	{		
      if( !(tmp & 0x4000) )					// nicht zu viele bits
        tmp <<= 1;							// shift
      if( !(rc5_bit & 1<<xRC5) )			// inverted bit
        tmp |= 1;							// insert new bit
      rc5_time = 0;							// zhle nchste pulse time
    }
  }

  rc5_tmp = tmp;
}
